import numpy as np
from typing import Any, Callable, Dict, List, Tuple

from .utils import DataPool


class ProblemHandler():
    """
    This class is used to provide a uniformed interface for the searching engine to
    interact with. It can store the searching space class, provide an entry to get a
    batch of scores from provided samples and hold all the searched results.

    :param BaseSpace search_space: Search space of handler.

    :Examples:

    >>> def target_func(x):
    >>>     ...
    >>> space = DiscreteSpace(5)
    >>> handler = ProblemHandler(space)
    >>> x = space.sample()
    >>> y = handler.get_score(np.array([x]), target_func)
    >>> x, y
    array([ 2 ]) array([ -2.25 ])
    >>> handler.get_all_data()
    array([[ 2 ]]) array([ -2.25 ])
    """

    def __init__(
            self,
            search_space: "BaseSpace",  # noqa
    ) -> None:
        self._search_space = search_space
        self._data_pool = DataPool(search_space)

    def get_score(self, new_samples: np.ndarray, target_function: Callable) -> np.ndarray:
        """
        Get scores of a batch of samples from target function. All samples and scores will be stored.

        :param np.ndarray new_samples: Samples to get scores.
        :param Callable target_function: Target function. The input and output should be one sample.
        :return np.ndarray: Scores of all provided samples.
        """
        scores = []
        for sample in new_samples:
            score = target_function(sample)
            scores.append(score)
        scores = np.array(scores)
        self._data_pool.update(new_samples, scores)
        return scores

    def get_cached_score(self, samples: np.ndarray) -> List:
        """
        Get scores of a batch of sample from stored scores. All the samples must be searched or updated
        before. Otherwise the score will be `None`.

        :param np.ndarray samples: Samples to get scores.
        :return List: Scores of all provided samples stored as `List` for there will be `None` elements.
        """
        return self._data_pool.get_score(samples)

    def get_all_data(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get all stores sample and scores.

        :return Tuple[np.ndarray, np.ndarray]: Stores samples and scores.
        """
        return self._data_pool.data

    def update_data(self, samples: np.ndarray, scores: np.ndarray) -> None:
        """
        Update a batch of samples and scores. If not stored already, they will be added into data pool.

        :param np.ndarray samples: Samples to be updated
        :param np.ndarray scores: Scores to be updated
        """
        assert samples.shape[0] == scores.shape[0], "sample and score must have the same num"
        self._data_pool.update(samples, scores)

    def provide_best(self) -> Tuple[np.ndarray, float]:
        """
        Get the currently best sample and its score in the data pool.

        :return Tuple[np.ndarray, float]: best sample and its score
        """
        if len(self) == 0:
            return
        return self._data_pool.best

    def __len__(self) -> int:
        return len(self._data_pool)

    @property
    def space(self) -> "BaseSpace":  # noqa
        return self._search_space

    @property
    def best_score(self) -> float:
        """
        Get the currently best score only.

        :return float: best score
        """
        if len(self._data_pool) == 0:
            return -np.inf
        return self._data_pool.best[1]

    def reset(self) -> None:
        """
        Reset the handler. Clear data pool.
        """
        self._data_pool.clear()



import gym
import copy
import numpy as np
from easydict import EasyDict
from typing import Any, Callable, Dict, List, Optional, Tuple

from .utils import DataPool
from ding.envs import BaseEnv, BaseEnvTimestep
from ding.utils.default_helper import deep_merge_dicts
from ding.utils import ENV_REGISTRY


@ENV_REGISTRY.register('problem_env')
class ProblemEnv(BaseEnv):

    config = dict(
        episode_len=float("inf"),
        init_type='zero',
        batch_num=1,
    )

    reward_space = gym.spaces.Box(low=0.0, high=1.0, shape=(1, ), dtype=np.float32)

    def __init__(
            self,
            cfg: Dict,
            search_space: "BaseSpace",  # noqa
            target_func: Callable,
    ) -> None:
        if 'cfg_type' not in cfg:
            self._cfg = self.__class__.default_config()
            self._cfg = deep_merge_dicts(self._cfg, cfg)
        else:
            self._cfg = cfg
        self._search_space = search_space
        self._obs_shape = self._search_space.shape
        self._data_pool = DataPool(search_space)
        self._target_func = target_func
        self._state = None

        self._episode_len = self._cfg.episode_len
        assert self._cfg.init_type in ['zero', 'random'], self._cfg.init_type
        self._batch_num = self._cfg.batch_num
        self._step_count = 0

    @property
    def ready_obs(self) -> Dict:
        obs = {}
        if self._state is None:
            return obs
        for env_id in range(self._batch_num):
            obs[env_id] = self._state[env_id]
        return obs

    @property
    def env_num(self) -> int:
        return self._batch_num

    def set_episode_len(self, episode_len: int) -> None:
        self._episode_len = min(episode_len, self._episode_len)

    def launch(self) -> None:
        self.reset()

    def reset(self, starting_point: Optional[np.ndarray] = None) -> None:
        self._step_count = 0
        if starting_point is None:
            if self._cfg.init_type == 'zero':
                single_obs = np.zeros(shape=self._obs_shape, dtype=np.float32)
            elif self._cfg.init_type == 'random':
                single_obs = self._search_space.sample()
        else:
            single_obs = starting_point
        self._state = np.stack([single_obs] * self._batch_num).astype(np.float32)

    def step(self, actions) -> Dict[int, BaseEnvTimestep]:
        env_ids = actions.keys()
        self._state = np.asarray(list(actions.values())).astype(np.float32)
        scores = np.apply_along_axis(self._target_func, axis=0, arr=self._state)
        self._data_pool.update(self._state, scores)
        self._step_count += 1
        done = self._step_count > self._episode_len
        timesteps = {}
        for env_id in env_ids:
            obs = self._state[env_id]
            reward = scores[env_id]
            info = {'final_eval_reward': scores[env_id]}
            timesteps[env_id] = BaseEnvTimestep(obs, reward, done, info)
        return timesteps

    def close(self) -> None:
        self._data_pool.clear()
        return

    def seed(self, seed: int, dynamic_seed: bool = True) -> None:
        self._seed = seed
        self._dynamic_seed = dynamic_seed
        np.random.seed(self._seed)

    @property
    def best(self) -> Tuple[np.ndarray, float]:
        return self._data_pool.best

    def __repr__(self) -> str:
        return super().__repr__()

    @classmethod
    def default_config(cls: type) -> EasyDict:
        cfg = EasyDict(copy.deepcopy(cls.config))
        cfg.cfg_type = cls.__name__ + 'Dict'
        return cfg



import torch
import torch.nn as nn
from typing import Any, Optional, Sequence, Union, Dict

from ding.model.common import ReparameterizationHead, DiscreteHead, MultiHead, DuelingHead, RegressionHead
from .rnn_model import LSTMSeqModel, GRUSeqModel


class DiscretePGModel(nn.Module):

    def __init__(
            self,
            obs_shape: int,
            action_shape: int,
            hidden_dim: int,
            sequence_len: int,
            num_layers: int = 1,
            model_type: str = 'lstm',
            dueling: bool = True,
            head_hidden_size: Optional[int] = None,
            head_layer_num: int = 1,
            activation: Optional[nn.Module] = nn.ReLU(),
            norm_type: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._obs_shape = obs_shape
        self._action_shape = action_shape
        self.input_layer = nn.Linear(self._obs_shape, hidden_dim * num_layers)
        assert model_type in ['lstm', 'gru'], model_type
        if model_type == 'lstm':
            self.encoder = LSTMSeqModel(hidden_dim, sequence_len, num_layers)
        elif model_type == 'gru':
            self.encoder = GRUSeqModel(hidden_dim, sequence_len, num_layers)

        head_hidden_size = hidden_dim

        if dueling:
            head_cls = DuelingHead
        else:
            head_cls = DiscreteHead
        multi_head = not isinstance(action_shape, int)
        if multi_head:
            self.head = MultiHead(
                head_cls,
                head_hidden_size,
                action_shape,
                layer_num=head_layer_num,
                activation=activation,
                norm_type=norm_type
            )
        else:
            self.head = head_cls(
                head_hidden_size, action_shape, head_layer_num, activation=activation, norm_type=norm_type
            )

    def forward(self, x: torch.Tensor) -> Dict:
        x = self.input_layer(x)
        x = x.unsqueeze(1)
        x, h = self.encoder(x, None)
        x = self.head(x)
        return x


class ContinuousPGModel(nn.Module):

    def __init__(
        self,
        obs_shape: int,
        action_shape: int,
        hidden_dim: int,
        sequence_len: int,
        action_space: str = 'regression',
        num_layers: int = 1,
        model_type: str = 'lstm',
        head_hidden_size: Optional[int] = None,
        head_layer_num: int = 1,
        activation: Optional[nn.Module] = nn.ReLU(),
        norm_type: Optional[str] = None,
        bound_type: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._obs_shape = obs_shape
        self._action_shape = action_shape
        self._action_space = action_space
        self.input_layer = nn.Linear(self._obs_shape, hidden_dim * num_layers)
        assert model_type in ['lstm', 'gru'], model_type
        if model_type == 'lstm':
            self.encoder = LSTMSeqModel(hidden_dim, sequence_len, num_layers)
        elif model_type == 'gru':
            self.encoder = GRUSeqModel(hidden_dim, sequence_len, num_layers)
        assert self._action_space in ['regression', 'reparameterization']
        if self._action_space == 'regression':
            self.head = RegressionHead(
                head_hidden_size,
                action_shape,
                head_layer_num,
                final_tanh=True,
                activation=activation,
                norm_type=norm_type
            )
        elif self._action_space == 'reparameterization':
            self.head = ReparameterizationHead(
                head_hidden_size,
                action_shape,
                head_layer_num,
                sigma_type='conditioned',
                activation=activation,
                norm_type=norm_type,
                bound_type=bound_type
            )

    def forward(self, inputs: Union[torch.Tensor, Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
        x = self.input_layer(inputs)
        x = x.unsqueeze(1)
        x, h = self.encoder(x, None)
        if self._action_space == 'regression':
            x = self.head(x)
            return {'action': x['pred']}
        elif self._action_space == 'reparameterization':
            x = self.head(x)
            return {'logit': [x['mu'], x['sigma']]}



import numpy as np
from typing import Any, Dict, Callable, List, Optional, Tuple

from .base_search_engine import BaseSearchEngine, ENGINE_REGISTRY
from disearch.problem import ProblemEnv
from ding.utils import deep_merge_dicts
from ding.worker import BaseLearner, SampleSerialCollector


class BaseRLEngine(BaseSearchEngine):

    config = dict(
        samples_per_iteration=128,
        max_iterations=100,
        env=dict(),
        policy=dict(collect=dict(collector=dict(), ), learn=dict(learner=dict(), )),
    )

    def __init__(
            self,
            cfg: Dict,
            search_space: "BaseSpace",  # noqa
            policy,
    ) -> None:
        cfg.policy.collect.collector = deep_merge_dicts(
            SampleSerialCollector.default_config(), cfg.policy.collect.collector
        )
        cfg.policy.learn.learner = deep_merge_dicts(BaseLearner.default_config(), cfg.policy.learn.learner)
        super().__init__(cfg, search_space)

        self._policy = policy
        self._collector = None
        self._learner = BaseLearner(cfg.policy.learn.learner, self._policy.learn_mode)
        self._replay_buffer = None
        self._best = None

        self._max_iterations = self._cfg.max_iterations
        self._samples_per_iteration = self._cfg.samples_per_iteration

    def reset(self) -> None:
        self._collector = None

    def propose(self, candidate_num: int) -> np.ndarray:
        new_data = self._collector.collect(candidate_num, train_iter=self._learner.train_iter)
        return new_data

    def search(self, target_func: Callable) -> Tuple[np.ndarray, int]:
        env = ProblemEnv(self._cfg.env, self._search_space, target_func)
        env.set_episode_len(self._samples_per_iteration // env.env_num)
        if self._collector is None:
            self._collector = SampleSerialCollector(self._cfg.policy.collect.collector, env, self._policy.collect_mode)
        else:
            self._collector.reset_env(env)

        for i in range(self._max_iterations):
            new_data = self.propose(self._samples_per_iteration)
            self.update_score(new_data)
            self._best = env.best
        return self.provide_best()

    def update_score(self, train_data: List) -> None:
        if self._replay_buffer is not None:
            self._replay_buffer.push(train_data, cur_collector_envstep=self._collector.envstep)
            for j in range(self._cfg.policy.learn.update_per_collect):
                train_data = self._replay_buffer.sample(self._cfg.policy.learn.batch_size, self._learner.train_iter)
                self._learner.train(train_data, self._collector.envstep)
            if self._learner.policy.get_attribute('priority'):
                self._replay_buffer.update(self._learner.priority_info)
        else:
            self._learner.train(train_data, self._collector.envstep)

    def provide_best(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._best



from abc import ABC, abstractmethod
from easydict import EasyDict
import copy
import numpy as np
from typing import Any, Callable, Dict

from ding.utils.registry import Registry
from ding.utils.default_helper import deep_merge_dicts


class BaseSearchEngine(ABC):

    config = dict()

    def __init__(self, cfg: Dict, search_space: "BaseSpace", random_state: Any = None) -> None:  # noqa
        if 'cfg_type' not in cfg:
            self._cfg = self.__class__.default_config()
            self._cfg = deep_merge_dicts(self._cfg, cfg)
        else:
            self._cfg = cfg

        self._search_space = search_space
        if random_state is None:
            self._random_state = np.random.RandomState()
        elif isinstance(random_state, int):
            self._random_state = np.random.RandomState(random_state)
        elif isinstance(random_state, np.random.RandomState):
            self._random_state = random_state
        else:
            raise TypeError(f'Unknown random state type - {repr(random_state)}.')

    @abstractmethod
    def reset(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def search(self, target_func: Callable):
        raise NotImplementedError

    @abstractmethod
    def propose(self, sample_num: int) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def update_score(self, samples: np.ndarray, scores: np.ndarray) -> None:
        raise NotImplementedError

    @abstractmethod
    def provide_best(self):
        raise NotImplementedError

    @classmethod
    def default_config(cls: type) -> EasyDict:
        cfg = EasyDict(cls.config)
        cfg.cfg_type = cls.__name__ + 'Config'
        return copy.deepcopy(cfg)


ENGINE_REGISTRY = Registry()



from __future__ import division
from __future__ import print_function

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
meta = {}
with open(os.path.join(here, 'disearch', '__init__.py'), 'r') as f:
    exec(f.read(), meta)

description = """DI-searching: """

setup(
    name=meta['__TITLE__'],
    version=meta['__VERSION__'],
    description=meta['__DESCRIPTION__'],
    long_description=description,
    author=meta['__AUTHOR__'],
    license='Apache License, Version 2.0',
    keywords='RL',
    packages=[
        *find_packages(include=('disearch', 'disearch.*')),
    ],
    python_requires='>=3.6',
    install_requires=[
        'di-engine>=0.3.1',
        'scikit-learn>=0.18.0',
        'protobuf<=3.20.1',  # for di-engine 0.3.1 bug
    ],
    extras_require={
        'doc': [
            'sphinx>=2.2.1',
            'sphinx_rtd_theme~=0.4.3',
            'enum_tools',
            'sphinx-toolbox',
        ],
        'test': [
            'pytest~=6.2.5',
            'pytest-cov~=3.0.0',
            'pytest-mock~=3.6.1',
            'pytest-xdist>=1.34.0',
            'pytest-rerunfailures~=10.2',
            'pytest-timeout~=2.0.2',
            'pytest-benchmark~=3.4.0'
        ],
        'style': [
            'yapf==0.29.0',
            'flake8',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],

)


import numpy as np
from typing import Any, Dict, Callable

from .base_search_engine import BaseSearchEngine, ENGINE_REGISTRY
from disearch.problem import ProblemHandler


@ENGINE_REGISTRY.register('random')
class RandomEngine(BaseSearchEngine):

    config = dict(num_sample=100, )

    def __init__(self, cfg: Dict, search_space: "BaseSpace", random_state: Any = None) -> None:  # noqa
        super().__init__(cfg, search_space, random_state)
        self._handler = ProblemHandler(search_space)

    def reset(self):
        self._handler.clear()

    def search(self, target_func: Callable):
        samples = self.propose(self._cfg.num_sample)
        self._handler.get_score(samples, target_func)
        return self.provide_best()

    def propose(self, sample_num: int) -> np.ndarray:
        samples = []
        for _ in range(sample_num):
            samples.append(self._search_space.sample())
        return np.array(samples)

    def update_score(self, samples: np.ndarray, scores: np.ndarray) -> None:
        self._handler.update_data(samples, scores)

    def provide_best(self):
        return self._handler.provide_best()



from abc import ABC, abstractclassmethod, abstractproperty
from math import ceil
from typing import Any, List, Type, Union

import numpy as np


class BaseSpace(ABC):
    r"""
    Abstract class of `space`.

    :param List shape: the shape of space, defaults to None
    :param Type dtype: the type of space, defaults to None
    :param Any random_state: the random method for space to generate sample, defaults to None
    """

    def __init__(self, shape: List = None, dtype: Type = None, random_state: Any = None) -> None:
        self._shape = None if shape is None else np.array(shape)
        self._dtype = None if dtype is None else np.dtype(dtype)
        if random_state is None:
            self._random_state = np.random.RandomState()
        elif isinstance(random_state, int):
            self._random_state = np.random.RandomState(random_state)
        elif isinstance(random_state, np.random.RandomState):
            self._random_state = random_state
        else:
            raise TypeError(f'Unknown random state type - {repr(random_state)}.')

    @abstractclassmethod
    def sample(self) -> np.ndarray:
        """
        Get a random sample from the space.

        :return np.ndarray: The random sample.
        """
        raise NotImplementedError

    @property
    def shape(self) -> np.ndarray:
        """
        Shape array of the space. It should be equal to the `shape` of a sample in space.

        :return np.ndarray: shape array.
        """
        return self._shape

    @property
    def dtype(self) -> Type:
        """
        Data type of the samples in searching space.

        :return Type: dtype.
        """
        return self._dtype


class DiscreteSpace(BaseSpace):
    """
    Discrete search space. It is configured by a input value or a list.
    The element of input at each position determinates the discrete
    space for each dim.

    :param List shape: list of space shape, defaults to None
    :param Type dtype: data type of samples, defaults to np.int64
    :param Any random_state: random state of sampler, defaults to None

    :Examples:

    >>> space = DiscreteSpace((2, 3, 4))
    >>> space.shape
    array([3, ])
    >>> x = space.sample()
    >>> x
    array([1, 1, 0])
    """

    def __init__(self, shape: List = None, dtype: Type = np.int64, random_state: Any = None) -> None:
        assert (np.array(shape) > 0).all(), "shape must be positive"
        if np.isscalar(shape):
            shape = [shape]
        self._nshape = np.asarray(shape, dtype=dtype)
        shape = self._nshape.shape

        super().__init__(shape, dtype, random_state)

    def sample(self) -> np.ndarray:
        sample = self._random_state.random_sample(self._nshape.shape) * self._nshape
        return sample.astype(self._dtype)

    def sample_single(self, position: int) -> int:
        """
        Get a random value at the position of the discrete space.

        :param int position: the position in the space to sample
        :return int: the random value at position
        """
        assert position < len(self._nshape)
        return self._random_state.randint(self._nshape[position])

    def __len__(self) -> int:
        return self._nshape.size

    @property
    def nshape(self) -> np.ndarray:
        """
        List of the space shape.

        :return np.ndarray: nshape
        """
        return self._nshape


class SeparateSpace(BaseSpace):
    """
    Separate search space. It is defined by a range of `start`, `end` and step to determine a list
    of values, and its dimension.

    :param int start: start value
    :param int end: end value, will NOT included in the space
    :param int step: step length, defaults to 1
    :param Union[int, List] dim: dimension or repeat times of the space, defaults to 1
    :param Type dtype: data type of space, defaults to np.int64
    :param Any random_state: random state for sampler, defaults to None

    :Examples:

    >>> space = SeparateSpace(start=1, end=6, step=2, dim=1)
    >>> space.shape
    array([1, ])
    >>> space.sample()  # must be in [1, 3, 5]
    1
    """

    def __init__(
            self,
            start: int,
            end: int,
            step: int = 1,
            dim: Union[int, List] = 1,
            dtype: Type = np.int64,
            random_state: Any = None
    ) -> None:
        self._start = start
        self._end = end
        self._step = step
        if np.isscalar(dim):
            dim = [dim]
        self._dim = np.array(dim)
        assert self._start >= 0 and self._end > 0 and self._step > 0 and (self._dim > 0).all(), \
            "all args must be positive!"
        self._count = ceil((self._end - self._start) / self._step)
        super().__init__(dim, dtype, random_state)

    def sample(self) -> np.ndarray:
        sample = self._random_state.randint(self._count, size=self._dim)
        sample = sample * self._step + self._start
        return sample.astype(self._dtype)

    def __len__(self) -> int:
        return self._count

    @property
    def nshape(self) -> np.ndarray:
        return self._shape


class ContinuousSpace(BaseSpace):
    """
    Continuous search space. The space is defined by its shape and lower/higher bounds, with continuous value.
    Users can set the lower and higher boundary of each dimension of the space, otherwise the value will be
    infinite at these dimensions. Each dimension can have different boundary types and values. If so, the shape
    of boundary argument must be the same as space's shape.

    :param List shape: shape of the space
    :param float low: the lower bound of space, defaults to None
    :param float high: the higher bound of space, defaults to None
    :param Type dtype: data type of space, defaults to np.float32
    :param Any random_state: random space of sampler, defaults to None

    :Example:

    >>> space = ContinuousSpace(shape=(2, 3), low=0, high=((5, 5, 5), (8, 8, 8)))
    >>> space.shape
    array([2, 3])
    >>> space.sample()
    array([[2.7791994, 4.216622 , 3.0280395],
       [4.2073464, 7.35141  , 2.5952444]], dtype=float32)
    """

    def __init__(
            self,
            shape: List,
            low: float = None,
            high: float = None,
            dtype: Type = np.float32,
            random_state: Any = None
    ) -> None:
        assert dtype is not None, "dtype must be explicitly provided."
        if np.isscalar(shape):
            shape = [shape]

        super().__init__(shape, dtype, random_state)

        if low is not None:
            assert (np.isscalar(low) or low.shape == self._shape), \
                "low.shape doesn't match provided shape"
            if np.isscalar(low):
                low = np.full(self._shape, low)
        else:
            low = np.full(self._shape, -np.inf)
        if high is not None:
            assert (np.isscalar(high) or high.shape == self._shape), \
                "low.shape doesn't match provided shape"
            if np.isscalar(high):
                high = np.full(self._shape, high)
        else:
            high = np.full(self._shape, np.inf)

        self._low = low.astype(dtype)
        self._high = high.astype(dtype)

        self._bounded_below = -np.inf < self._low
        self._bounded_above = np.inf > self._high

    def sample(self) -> np.ndarray:
        """
        Get a random sample from the space. unbounded dimension will be sampled from a normal distribution,
        dimensions with one bound will be sampled from a exponential distribution. Others will be sampled
        from a exponential distribution.

        :return np.ndarray: The random sample.
        """
        sample = np.empty(self._shape)

        unbounded = ~self._bounded_below & ~self._bounded_above
        upp_bounded = ~self._bounded_below & self._bounded_above
        low_bounded = self._bounded_below & ~self._bounded_above
        bounded = self._bounded_below & self._bounded_above

        sample[unbounded] = np.random.normal(size=unbounded[unbounded].shape)
        sample[low_bounded] = (np.random.exponential(size=low_bounded[low_bounded].shape) + self._low[low_bounded])

        sample[upp_bounded] = (-np.random.exponential(size=upp_bounded[upp_bounded].shape) + self._high[upp_bounded])

        sample[bounded] = np.random.uniform(
            low=self._low[bounded], high=self._high[bounded], size=bounded[bounded].shape
        )

        return sample.astype(self._dtype)

    def __len__(self) -> int:
        return np.prod(self._shape)

    @property
    def nshape(self) -> np.ndarray:
        return self._shape

    @property
    def bounds(self) -> np.ndarray:
        return np.array(list(zip(self._low, self._high)))


class HybridSpace(BaseSpace):
    """
    Hybrid searching space consists of one or more kinds of space above.

    :param space_kwargs: the provided space list

    :Example:

    >>> s1 = DiscreteSpace(3)
    >>> s2 = ContinuousSpace(2)
    >>> space = HybridSpace(s1, s2)
    >>> space.shape
    [array([1]), array([2])]
    >>> space.sample()
    [array(1), array([-0.35141  , 1.5952444], dtype=float32)]
    """

    def __init__(self, *space_kwargs) -> None:
        self._spaces = []
        for space in space_kwargs:
            self._spaces.append(space)

    def sample(self) -> np.ndarray:
        """
        Get a list of random samples from each space.

        :return np.ndarray: sample list
        """
        samples = np.zeros(shape=(len(self, )), dtype=object)
        for i in range(len(self)):
            space = self._spaces[i]
            samples[i] = space.sample()
        return np.array(samples, dtype=object)

    def __len__(self) -> int:
        """
        Get the num of spaces.

        :return int: space num
        """
        return len(self._spaces)

    @property
    def shape(self) -> List:
        """
        Get a shape list of each space.

        :return List: space list
        """
        shape = []
        for space in self._spaces:
            shape.append(space.shape)
        return shape

    @property
    def nshape(self) -> np.ndarray:
        """
        get a nshape list of each space.

        :return np.ndarray: nshape list
        """
        nshape = []
        for space in self._spaces:
            nshape.append(space.nshape)
        return np.array(nshape)



from typing import Tuple
import numpy as np

from .space import HybridSpace


def make_hashable(x):
    if isinstance(x, np.ndarray):
        if x.dtype == object:
            x_ = x.copy()
            x = []
            for xi in x_:
                x += xi.flatten().tolist()
        else:
            x = x.flatten()
    return tuple(map(float, x))


class DataPool():

    def __init__(self, search_space: "BaseSpace") -> None:  # noqa
        if isinstance(search_space, HybridSpace):
            self._samples = np.empty(shape=(0, len(search_space)))
        else:
            self._samples = np.empty(shape=(0, *search_space.shape))
        self._scores = np.empty(shape=(0, ))
        self._best_idx = None
        self._sample_dict = {}

    def update(self, samples: np.ndarray, scores: np.ndarray) -> None:
        new_idx = []
        for i in range(samples.shape[0]):
            sample, score = samples[i], scores[i]
            if make_hashable(sample) not in self._sample_dict:
                self._sample_dict[make_hashable(sample)] = score
                new_idx.append(i)
        self._samples = np.concatenate([self._samples, samples[new_idx]], axis=0)
        self._scores = np.concatenate([self._scores, scores[new_idx]], axis=0)
        self._best_idx = np.argmax(self._scores)

    def get_score(self, samples: np.ndarray) -> np.ndarray:
        scores = []
        for sample in samples:
            try:
                score = self._sample_dict[make_hashable(sample)]
            except KeyError:
                print(f"sample {sample} not found in cache!")
                score = None
            scores.append(score)
        return scores

    def __len__(self) -> int:
        return len(self._sample_dict)

    def clear(self) -> None:
        self._searched_samples = np.empty(shape=(0, *self._samples.shape[1:]))
        self._searched_scores = np.empty(shape=0)
        self._sample_dict = {}
        self._best_idx = None

    @property
    def data(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._samples, self._scores

    @property
    def best(self) -> Tuple[np.ndarray, float]:
        if self._best_idx is None:
            raise ValueError("sample pool is empty!")
        return self._samples[self._best_idx], self._scores[self._best_idx]



import numpy as np
import warnings
from collections import deque
from typing import Callable, Dict, Any, Tuple
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern
from scipy.optimize import minimize
from scipy.stats import norm

from .base_search_engine import BaseSearchEngine, ENGINE_REGISTRY
from disearch.problem import ProblemHandler


def ucb(x, gp, kappa):
    mean, std = gp.predict(x, return_std=True)
    return mean + kappa * std


def ei(x, gp, y_max, xi):
    mean, std = gp.predict(x, return_std=True)
    a = (mean - y_max - xi)
    z = a / std
    return a * norm.cdf(z) + std * norm.pdf(z)


def poi(x, gp, y_max, xi):
    mean, std = gp.predict(x, return_std=True)
    z = (mean - y_max - xi) / std
    return norm.cdf(z)


@ENGINE_REGISTRY.register('bo')
class BayesianOptimizationEngine(BaseSearchEngine):
    """
    The searching engine of Bayesian Optimization algorithm. The engine takes a function to optimize
    in a discrete space as well as the bounds in order to find the value for samples yield the maximum
    targets using bayesian optimization.

    :param Dict cfg: user config
    :param BaseSpace search_space: searching space of engine
    """

    config = dict(
        init_point=5,
        max_iterations=25,
        acquisition='ucb',
        kappa=2.576,
        kappa_decay=1,
        kappa_decay_delay=0,
        xi=0.0,
        verbose=False,
    )

    def __init__(
            self,
            cfg: Dict,
            search_space: "BaseSpace",  # noqa
            random_state: Any = None,
    ) -> None:
        super().__init__(cfg, search_space, random_state)

        self._gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            alpha=1e-6,
            normalize_y=True,
            n_restarts_optimizer=5,
            random_state=self._random_state,
        )
        self._handler = ProblemHandler(search_space)

        self._queue = deque()

        self._init_point = self._cfg.init_point
        self._max_iterations = self._cfg.max_iterations
        self._kappa = self._cfg.kappa
        self._kappa_decay = self._cfg.kappa_decay
        self._kappa_decay_delay = self._cfg.kappa_decay_delay
        self._xi = self._cfg.xi
        assert self._cfg.acquisition in ['ucb', 'ei', 'poi'], self._cfg.acquisition
        self._acquire_type = self._cfg.acquisition
        self._verbose = self._cfg.verbose

    def reset(self) -> None:  # noqa
        self._handler.clear()
        self._queue.clear()
        self._gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            alpha=1e-6,
            normalize_y=True,
            n_restarts_optimizer=5,
            random_state=self._random_state,
        )

    def _predict(self, x, y_max):
        if self._acquire_type == 'ucb':
            return ucb(x, self._gp, self._kappa)
        elif self._acquire_type == 'ei':
            return ei(x, self._gp, y_max, self._xi)
        elif self._acquire_type == 'poi':
            return poi(x, self._gp, y_max, self._xi)
        else:
            raise ValueError('Unknown aquisition type:', self._acquire_type)

    def _acquisition_max(self, n_warmup=10000, n_iter=10):
        x_tries = np.array([self._search_space.sample() for _ in range(n_warmup)])
        y_max = self._handler.best_score
        ys = self._predict(x_tries, y_max)
        x_max = x_tries[ys.argmax()]
        max_acq = ys.max()

        x_seeds = np.array([self._search_space.sample() for _ in range(n_iter)])
        for x_try in x_seeds:
            res = minimize(
                lambda x: -self._predict(x.reshape(1, -1), y_max=y_max),
                x_try.reshape(1, -1),
                bounds=self._search_space.bounds,
                method="L-BFGS-B"
            )

            if not res.success:
                continue

            try:
                fun = -res.fun[0]
            except TypeError:
                fun = -res.fun

            if max_acq is None or -fun > max_acq:
                x_max = res.x
                max_acq = fun

        return np.clip(x_max, self._search_space.bounds[:, 0], self._search_space.bounds[:, 1])

    def search(self, target_func: Callable) -> Tuple[np.ndarray, float]:  # noqa
        init_point = self._init_point
        if len(self._queue) == 0 and len(self._handler) == 0:
            init_point = max(init_point, 1)
        for i in range(init_point):
            sample = self._search_space.sample()
            self._queue.append(sample)

        iterations = 0
        while len(self._queue) > 0 or iterations < self._max_iterations:
            if len(self._queue) > 0:
                sample = self._queue.pop()
            else:
                sample = self.propose()
                iterations += 1
            sample = sample.reshape(1, -1)
            score = self._handler.get_score(sample, target_func)
            self.update_score(sample, score)
            best_sample, best_score = self._handler.provide_best()
            if not self._verbose:
                print(" Iteration {} best score: {:.4f} .".format(iterations + 1, best_score))
        return best_sample, best_score

    def propose(self, sample_num: int = 1) -> np.ndarray:
        sample_list = []
        while len(sample_list) < sample_num:
            if len(self._queue) > 0:
                candidate = self._queue.pop()
            else:
                candidate = self._acquisition_max()
            sample_list.append(candidate)
        return np.array(sample_list)

    def update_score(self, sequences: np.ndarray, scores: np.ndarray) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._gp.fit(sequences, scores)

    def provide_best(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._handler.provide_best()



import torch
import torch.nn as nn
from typing import Optional, Tuple


class LSTMSeqModel(nn.Module):

    def __init__(
            self,
            hidden_dim: int,
            sequence_len: int,
            num_layers: int = 1,
    ) -> None:
        super().__init__()
        self._hidden_dim = hidden_dim
        self._seq_len = sequence_len
        self._num_layers = num_layers

        self._rnn = nn.LSTM(hidden_dim, hidden_dim, num_layers, batch_first=True)

    def _init_hidden_state(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        seq_len = inputs.shape[1]
        batch_size = inputs.shape[0]
        zeros = torch.zeros(self._num_layers, batch_size, self._hidden_dim, dtype=inputs.dtype, device=inputs.device)
        return zeros, zeros

    def forward(self, inputs: torch.Tensor, prev_state: Optional[Tuple]) -> Tuple[torch.Tensor, Tuple]:
        if prev_state is None:
            prev_state = self._init_hidden_state(inputs)
        output = []
        x = inputs
        h = prev_state
        for _ in range(self._seq_len):
            x, h = self._rnn(x, h)
            output.append(x)
        output = torch.cat(output, dim=1)
        return output, h


class GRUSeqModel(nn.Module):

    def __init__(
        self,
        hidden_dim: int,
        sequence_len: int,
        num_layers: int = 1,
    ):
        super().__init__()
        self._hidden_dim = hidden_dim
        self._seq_len = sequence_len
        self._num_layers = num_layers

        self._rnn = nn.GRU(hidden_dim, hidden_dim, num_layers)

    def _init_hidden_state(self, inputs: torch.Tensor) -> torch.Tensor:
        seq_len, batch_size = inputs.shape[:2]
        zeros = torch.zeros(self._num_layers, batch_size, self._hidden_dim, dtype=inputs.dtype, device=inputs.device)
        return zeros

    def forward(self, inputs: torch.Tensor, prev_state: Optional[torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        if prev_state is None:
            prev_state = self._init_hidden_state(inputs)
        output = []
        x = inputs
        h = prev_state
        for _ in range(self._seq_len):
            x, h = self._rnn(x, h)
            output.append(x)
        output = torch.cat(output, dim=1)
        return output, h



import numpy as np
import copy
from typing import Any, Dict, Sequence, Tuple

from .base_search_engine import BaseSearchEngine, ENGINE_REGISTRY
from disearch.problem import ProblemHandler


@ENGINE_REGISTRY.register('ga')
class GeneticAlgorithmEngine(BaseSearchEngine):

    config = dict(
        cross_rate=0.1,
        mutate_rate=0.1,
        max_generation=100,
        population_size=100,
        verbose=False,
    )

    def __init__(self, cfg: Dict, search_space: "BaseSpace", random_state: Any = None) -> None:  # noqa
        super().__init__(cfg, search_space, random_state)
        self._cross_rate = self._cfg.cross_rate
        self._mutate_rate = self._cfg.mutate_rate

        self._handler = ProblemHandler(search_space)
        self._first_generation = True
        self._verbose = self._cfg.verbose

    def reset(self) -> None:
        self._handler.clear()
        self._first_generation = True

    def search(self, target_func) -> Tuple[np.ndarray, float]:
        for i in range(self._cfg.max_generation):
            pop = self.propose(self._cfg.population_size)
            score = self._handler.get_score(pop, target_func)
            self.update_score(pop, score)
            best_seq, best_score = self.provide_best()
            if not self._verbose:
                print(" Generation {} best score: {:.3f} .".format(i + 1, best_score))
        return best_seq, best_score

    def propose(self, candidate_num: int) -> np.ndarray:
        if self._first_generation:
            candidates = np.vstack([self._search_space.sample() for _ in range(candidate_num)])
            self._first_generation = False
        else:
            samples, scores = self._handler.get_all_data()
            parents_idx = self._random_state.choice(
                np.arange(len(self._handler)), candidate_num, True, p=scores / np.sum(scores)
            )
            parents = samples[parents_idx]
            children = self._crossover(parents, parents.copy())
            children = self._mutation(children)
            candidates = np.vstack(children)
        return candidates

    def _crossover(self, seqs: np.ndarray, pops: np.ndarray) -> np.ndarray:
        for i in range(len(seqs)):
            if self._random_state.rand() < self._cross_rate:
                j = self._random_state.randint(0, len(seqs), size=1)
                cross_points = self._random_state.randint(0, 2, len(self._search_space)).astype(bool)
                seqs[i, cross_points] = pops[j, cross_points]
        return seqs

    def _mutation(self, seqs: np.ndarray) -> np.ndarray:
        for seq in seqs:
            for point in range(len(self._search_space)):
                if self._random_state.rand() < self._mutate_rate:
                    seq[point] = self._random_state.randint(self._search_space.nshape[point])
        return seqs

    def update_score(self, sequences: np.ndarray, scores: np.ndarray) -> None:
        self._handler.update_data(sequences, scores)

    def provide_best(self) -> Tuple[np.ndarray, float]:
        if self._first_generation:
            return
        return self._handler.provide_best()
